VDK-PYLIB
=========

vdk-pylib is a MKit-based toolkit providing templates, plugins and build
targets specific for simple Python libraries.

If you are unfamiliar with MKit, here's a reminder of which targets
it provides you with:

 *  Basic targets such as `make build`, `make clean`, `make dist`
    and `make install`,

 *  `make vbump` and `make release` for SemVer-compliant version
    management and pre-release operations.

With vdk-pylib, you will get following targets ready to go:

 *  `make pystuff` for building PyPI distribution,

 *  `make debstuff` for building Debian source files (tarball and
    *debian* directory),

 *  `make rpmstuff` for building RPM source files (tarball and
    *.spec* file)

 *  `make vdk_test` for quick-building and running test suites
    directly from your project's git directory.

    This target supports MyPy, doctests, standard Python's unittest suites
    (pytest support is on the horizon) as well as possibility to include
    own suites.

Technically, the first three targets (`*stuff`) are already provided
by MKit, but they require you to create and maintain templates for each
target (setup.py, template.spec and debian templates).

vdk-pylib provides these templates out-of-the-box, allowing you to
focus on your own code while always having the ability to test your code
in relevant environments such as PyPI, venv, Debian-based or Fedora-based
distributions, or all at once.

Furthermore, vdk-pylib can be especially useful if you have multiple (or
many) libraries that you want to maitain separately.  Using vdk-pylib
for all of your libraries will add consistency and predictability to
your development process, simplifying testing, releases or CI pipeline
scripting.


Limitations
-----------

First, vdk-pylib does not aim to support full set of features of Python
build systems.  Furthermore, vdk-pylib is itself in early development
stage.

Features not supported at this stage include:

 *  entrypoints and Python's standard plugin system,
 *  advanced requirements,
 *  distribution of non-Python files (it might be already
    supported, though),

Features which are likely not going to be supported include:

 *  packages using Python C/C++ API,
 *  more advanced and complex build scenarios.

On the other hand, if your Python library consists only of pure Python
files, vdk-pylib can be a quick way to greatly speed up your development.

Second, while vdk-pylib helps you create usable Debian or Fedora packages,
these are by no means meant to be submitted to official repositories.
This is because meta-data and scripting (meaning .spec file and debian
directory) generated by vdk-pylib does not follow official guidelines or
best practices.

On the other hand, vdk-pylib-driven packages can be suitable for testing,
in well-defined environments (such as your own maintained cloud) or for
project in early stages of development.

Finally, both MKit and vdk-pylib require the project to be maintained
by Git.  There are no plans to support other versioning systems.


Getting started
---------------

Enabling vdk-pylib for your project is done by installing vdk-pylib package and
adding mkit.ini file and a Makefile to your project.  These files will use include
directives to enable templates, configuration and build targets provided by
vdk-pylib.


### Installing vdk-pylib ###


vdk-pylib is supported only on GNU/Linux distributions.


#### Pre-built packages - Fedora ####

Fedora, EPEL and CentOS Stream packages are available in `netvor/vornet-infra`
COPR repos:

    sudo dnf copr enable netvor/vornet-infra
    sudo dnf install vdk-pylib


#### Pre-built packages - Debian ####

Debian 12 packages are available in custom repo at Vornet Scratch APT
repo; [follow instructions][apt] to enable the APT repository

  [apt]: https://apt.vornet.cz/vscratch/

After enabling the repository, you can install vdk-pylib as any other
package:

    sudo apt install vdk-pylib


#### Installing manually ###

***Warning**: manual install is not recommended for normal usage,
since there is no guarrantee that uninstalling or upgrading vdk is
going to work as expected.  Manual installation should normally used
only in isolated environment such as docker containers or when building
distribution packages.*

vdk-pylib depends on `make`, `git`, `python3-build`, `python3-venv` and
other tools usually available on most GNU/Linux distributions, such
as GNU core utils and Bash.

After satisfying the dependencies, you can install vdk-pylib:

    sudo make install

This command will respect PREFIX and DESTDIR variables.  Default values
are `PREFIX=/usr/local` and `DESTDIR=/`.  The default PREFIX conforms
to FHS standard, but examples used here will not work with it, since
they expect `/usr` prefix.

In places like Docker container, you probably want to use `PREFIX=/usr/local`.


### mkit.ini ###

This file contains all of your basic project meta-data.  Here's an example
of mkit.ini for a simple Python module called *bakebot* which depends on
another module called *kitchen*.

    [INCLUDE:/usr/share/vdk/pylib/skel/mkit.ini]

    [project]
        version     = 0.1.2
        maintainer  = Joe Maintainer <joe@example.com>
        name        = BakeBot
        tagline     = Baking library
        vcs_browser = https://gitlab.example.com/joe/[project:pkgname]

    [macros]
        __VDK_PYLIB_MODNAME__       = bakebot
        __VDK_PYPI_REQUIRES__       = kitchen
        __VDK_DEBIAN_DEPENDS__      = python3-kitchen
        __VDK_FEDORA_REQUIRES__     = python3-kitchen
        __VDK_PYLIB_SUMMARY__       = [project:tagline]
        __VDK_PYLIB_DESCRIPTION__   = This module helps you with baking.
        __VDK_PYLIB_DESCRIPTION__   =
        __VDK_PYLIB_DESCRIPTION__   = Most people don't know they can
        __VDK_PYLIB_DESCRIPTION__   = bake with Python, but now you can!

    #mkit version=0.1.2

Typically you only need to specify this file once.

The mkit.ini file is using an INI syntax dialect and should be mostly
intuitive.  Only few things are worth noting syntax-wise:

 *  multi-line values are specified by repeating the same key for
    each line (this can include empty lines),

 *  values can reference other values by including special
    syntax `[<SECTION>:<KEY>]`

 *  there is no need for quoting.

In this example, aside from from self-descriptive fields:

 *  `__VDK_PYLIB_MODNAME__` specifies Python name of the module.

 *  `__VDK_PYPI_REQUIRES__` specifies Python name of another
    module which this module depends on (as in `pip install kitchen`)

 *  `__VDK_FEDORA_REQUIRES__` and `__VDK_DEBIAN_DEPENDS__` specify Fedora and
    Debian package names (as in `dnf install python3-kitchen` or
    `apt install python3-kitchen`)

See below for a more complete description of these field as well as other
fields.
    

### Makefile ###

The Makefile only needs to contain single line:

    include /usr/share/vdk/pylib/Makefile

and normally needs to be touched again.  Adding this file will enable you to use
`make` command to interact with MKit's built-in targets as well as vdk-pylib's
targets and templates.


### _meta.py.skel (optional) ###

For your convenience, vdk-pylib provides template of `_meta.py.skel`, which
you can include in your project.  vdk-pylib will turn this file to `_meta.py`
so that you can access static as well as build-time meta-data from your program:

    from ._meta import NAME, VERSION, MAINTAINER
    print(f"This is {NAME} v{VERSION}; please report bugs to {MAINTAINER}")

See below for the full content of the `_meta.py.skel` template.


### Checking your configuration ###

After including the mentioned files and commiting them to the Git repository,
you can review your configuration by running

    make _mkit_show_metadata

You should see a bunch of macros.  Some of them correspond to, or are derived
from values you have specified, some are based on Git info, and others are
built-in defaults provided by vdk-pylib.  Most macros are described below,
some are self-descriptive.

Take a quick look if they seem correct.


Usage
-----

### Producing build assets ###

Once basic meta-data is configured, you should be able to start with running

    make pystuff

Among other files, this will produce

 *  `setup.py` script,
 *  `dist` directory,
 *  `bakebot.egg-info` directory,
 *  `mkvenv.sh` script,

`mkvenv.sh` is a simple Bash script which you can run to create virtual
environment which includes your module as well as its dependencies.

Rest of the assets are standard Python build assets.  `setup.py` has
been already ran to create the ther files.  `dist` contains files
for releasing the package (eg. to PyPI).

You can clean up these files for now:

    make clean


### Testing ###

At any moment, you can re-build and run test suite:

    make vdk_test

This will detect test suites under *tests* directory, if you have some.
See below for further details on how to make sure your tests are detected
and how to configure behavior of the target such as cleanup, early exit,
test arguments or selection of specific test suites.

Another quick & easy to run command is

    make vdk_test_shell

This will re-build your project and run Bash interpreter with PYTHONPATH
set so that you can run eg. ipython3 and see how your module behaves.
Note that this opens new Bash process, so you need to run `exit` or
press `Ctrl+D` to clean up and get back to your shell.


### Versioning ###

vdk-pylib will construct full and unique version for any branch or commit.

In most cases, the version will not be a "clean" release version, instead
it should look like:

    <BASE>+t<TIME>.<BRANCHNAME>.g<GIT_SHORTHASH>

where

 *  `<BASE>` is the base version specified in your *mkit.ini* file.

 *  `<TIME>` is commit time of the latest commit,

 *  `<BRANCHNAME>` is current branch name,

 *  `<GIT_SHORTHASH>` is short SHA1 of the last commit.

Occasionally, the version might contain following suffixes:

 *  `.dirty` if files have been changed since the last commit,

 *  `.async` if current branch is set to follow a remote branch,
    but the current branch is ahead the remote branch.  (That is,
    this version might be obsolete.)

For eample, a string might look like:

    1.12.0+t20240329151734.foo.g2f2c6a1

meaning it has "base" version of `1.12.0`, and was built from
branch `foo`, where at the moment, the last commit was created
March 29th, 2024, 15:17 (note that this is *AuthorDate*, not
*CommitDate*; see below for details) and its short SHA1 was
`2f2c6a1`.

This version string is used in all assets, from the mentioned
`_meta.py`, `setup.py` and all RPM or DEB files.

Furthermore, the version is constructed in such a way that it *sorts
properly*.  In other words, if you create new commit, then create packages
assets and try to install them *over* older ones, most package managers
such as *rpm* or *deb* will recognize this as upgrade, and permit it
without any further interaction.


Releases
--------

Fully-qualified version are useful for development and testing,
but for official releases you might want o use clean versions
such as `1.12.1`.

vdk-pylib will generate clean version under following circumstances:

 *  you are at "release source" branch, whihc by default is `main`,

 *  the last commit contains tag which is exactly `v<VERSION>`, where
    `<VERSION>` is identical to version currently specified in
    *mkit.ini*.


### MKit release workflow ###

The above circumstances are usually true if you follow MKit's release
workflow:

 1. Once you are ready to release, create "bump commit", which is a
    special commit which does not touch any code -- it only increases
    version in mkit.ini and serves as a checkpoint in the commit
    history with a human-readable summary of changes (changelog).

    You can do this by running one of following commands, depending
    on which version component you want to bump (see SemVer for
    details):

     *  `make vbump_x` for major release (eg. 1.2.3 -> 2.0.0)
     *  `make vbump_y` for minor release (eg. 1.2.3 -> 1.3.0)
     *  `make vbump_z` for patch release (eg. 1.2.3 -> 1.2.4)
     *  `make vbump` (identical as `vbump_z`)

    Each of this command will:

     1. Check if there are any commits beginning with `WIP`

     2. Bump the `version` field in `[project]` secton of mkit.ini.

     3. Generate simple list of commits since the last clean version.

     4. Open your editor with git commit message draft (see git-commit
        manpage), where you are expected to create the human-readable
        summary.

        The summary should summarize changes and serve as alternative
        to reading the actual commit log.  (That is to say, don't just
        repeat commit log here!)

     4. Once you're done and closed the message file, the commit will
        be created.

 2. Create the actual release

    Run one of commands

     *  `make release_x` for major release (eg. 1.2.3 -> 2.0.0)
     *  `make release_y` for minor release (eg. 1.2.3 -> 1.3.0)
     *  `make release_z` for patch release (eg. 1.2.3 -> 1.2.4)
     *  `make release` (identical as `vbump_z`)

    Each of this command will

     1. Check if the version in mkit.ini has been updated.

     2. Create *release tag*, i.e. git tag in format `v<VERSION>`
        (eg. `v1.3.0`) with the summary from the bump commit
        attached.

     3. If you also configured "release destination" branch, it
        will update it to point to current release.

 3. Push all tags and branches.

    Now your version is released.

Note that immediately after `make vbump` and `make release` and before
adding more commits, any other targets will use clean version in their
meta-data.

Overall, this workflow is desinged to allow reliable and transparent
versioning on single linear commit history, without much fuss with
separate release branches or merging back and forth.



Further configuration
---------------------


### mkit.ini -- basic fields ###

Full set of fields is available as part of MKit documentation; here
we'll only describe most common fields.

Basic meta-data (`[project]` section):

 *  `version` - SemVer 2.0 compliant version.

    This should almost never be edited manually.  See *Versioning* and
    *Releases* sections to understand how this is used.

 *  `maintainer` - maintainer contact in form of `Name <email@example.com>`

 *  `name` - Short human-oriented name of the project

 *  `tagline` - Single lined description of the project

 *  `vcs_browser` - URL of a page where source code can be
    viewed.

    With systems like GitHub and GitLab this is commonly
    also git clone URL as well as main project homepage.

`[macros]` section contains vdk-pylib specific values.


### mkit.ini -- vdk-pylib macros ###

Required macros (`[macros]` section):

 *  `__VDK_PYLIB_MODNAME__`

    Python module name. (as in `

Other supported macros:

 *  `__VDK_PYLIB_SUMMARY__`

    Single-line package description.  By default:

        [project:name] - [project:tagline]

    eg.

        BakeBot - Baking library

 *  `__VDK_PYLIB_DESCRIPTION__`

    Library description; may be multiple lines.

 *  `__VDK_DEBIAN_DEPENDS__`, `__VDK_DEBIAN_BUILD_DEPENDS__`

    Names of required Debian packages.  Used to create `Depends:`
    and `Build-Depends:` fields in the debian/control file, respectively.

    Beware: some tests we run tests during packaging (debian/rules) will
    require package's runtime dependencies.  Namely mypy, and depending on
    your code, also pytest and/or unittest might require your dependencies
    to be present.  Since vdk-pylib-0.1.3, at least mypy is mandatory.

    Therefore `__VDK_DEBIAN_DEPENDS__` is automatically included into
    Build-Depends.  This might make testing more expensive and even potentially
    overshadow some issues such as missing build dependencies but given
    VDK's main use case (early and unofficial development) this is an
    acceptable trade-off.  Currently this behavior cannot be configured.

    By default, `python3` is included in both fields.

    Despite the name, package names don't strictly need to be present in
    Debian; if you maintain your own repository or use a Debian-derived
    distribution such as Ubuntu or Linux Mint, these macros will work
    as long as you enable all necessary repositories before installing
    or building your package.

 *  `__VDK_FEDORA_REQUIRES__`, `__VDK_FEDORA_BUILDREQUIRES__`

    Names of required Fedora packages.  Used to create `Requires:`
    and `BuildRequires:` fields in the RPM .spec file, respectively.

    Beware: some tests we run tests during packaging (`%check` section) will
    require package's runtime dependencies.  Namely mypy, and depending on
    your code, also pytest and/or unittest might require your dependencies
    to be present.  Since vdk-pylib-0.1.3, at least mypy is mandatory.

    Therefore `__VDK_FEDORA_REQUIRES__` is automatically included into
    BuildRequires.  This might make testing more expensive and even potentially
    overshadow some issues such as missing build dependencies but given
    VDK's main use case (early and unofficial development) this is an
    acceptable trade-off.  Currently this behavior cannot be configured.

    By default, `python3` is included in `__VDK_FEDORA_REQUIRES__`
    and `python3-devel` and `python3-setuptools` in
    `__VDK_FEDORA_BUILDREQUIRES__`.

    Despite the name, package names don't strictly need to be present in
    Fedora; if you maintain your own repository (eg. using Fedora COPR)
    or use a Fedora-derived distribution such as CentOS or RHEL, these
    macros will work as long as you enable all necessary repositories
    before installing or building your package.

 *  `__VDK_PYPI_REQUIRES__`

    Names of required Python modules.  Used to form `requires` keyword
    to distutils.core.setup() in setup.py.

    Despite the name, package names don't strictly need to be present
    on PyPI; if you maintain your own repository, this macro will work
    as long as you enable all necessary repositories before installing
    your package.

 *  `__VDK_PYLIB_PKGMAP__`

    Contains mapping of packages and their directories and files.  Each
    line must be in this form:

        PKGNAME  PKGDIR   [FILE]..

    where PKGNAME is package name including any embedded dots, PKGDIR
    is path to the directory relative to 'src' directory, and each FILE
    is path to package's data file relative to the PKGDIR.

    For example, if a package 'foo' has sub-packages 'bar' and 'baz',
    where ther latter has a data files baz.db and baz.json, package
    map could look like this:

       __VDK_PYLIB_PKGMAP__ = foo     src/foo
       __VDK_PYLIB_PKGMAP__ = foo.bar src/foo/bar
       __VDK_PYLIB_PKGMAP__ = foo.baz src/foo/baz baz.db baz.json

    If __VDK_PYLIB_PKGMAP__ is omitted, only one line is included, assuming
    that only one Python package is produced, named `__VDK_PYLIB_MODNAME__`
    and it contains just a single file named 'py.typed'.

       __VDK_PYLIB_PKGMAP__ = foo     src/foo     py.typed

    This map is used to construct fields `packages`, `package_dir` and
    `package_data` in *setup.py*.

 *  `__VDK_PYLIB_CLASSIFIERS__`, `__VDK_PYLIB_CLASSIFIERS_OVERRIDE__`

    Both macros specify list of PyPI classifiers to add to package meta-data.

    Use `__VDK_PYLIB_CLASSIFIERS__` if you want to add to the default set
    of classifiers; use `__VDK_PYLIB_CLASSIFIERS_OVERRIDE__` if you want
    to override it and define your own set of classifiers.

    Final set of classifiers is built like this:

     1. Classifiers derived from `__VDK_PYLIB_PYTHON_GE__` and
        `__VDK_PYLIB_PYTHON_LT__` are always included.

        This also always includes `Programming Language :: Python`
        and `Programming Language :: Python :: 3 :: Only`.

     2. If defined, `__VDK_PYLIB_CLASSIFIERS_OVERRIDE__` are
        added and the set is considered complete.

     3. If, and only if the above macro is *not* defined,
        `__VDK_PYLIB_CLASSIFIERS__` is used instead.

    The reason two macros are needed is that mkit.ini format only allows
    additive definition, meaning that by defining the macro you can only
    *add* items to default values.  To minimize boilerplate, VDK includes
    reasonable set of classifiers already, but the mkit.ini syntax does
    not allow you to remove items from it.

    We get the best of both worlds by having VDK's setup.py choose
    between the two macros on run-time.  This makes it easy to
    add to the defaults by `__VDK_PYLIB_CLASSIFIERS__`, but when necessary,
    allows full control by `__VDK_PYLIB_CLASSIFIERS_OVERRIDE__`.

 *  `__VDK_PYLIB_PYTHON_GE__`, `__VDK_PYLIB_PYTHON_LT__`

    Specifies lower and upper bounds of Python compatible versions.  These
    will be used to add PyPI classifiers to the package meta-data.

    Default values are 3.9 and 3.13, respectively.

    `__VDK_PYLIB_PYTHON_GE__` specifies the first compatible version,
    `__VDK_PYLIB_PYTHON_LT__` specifies the first non-compatible version.

    For example:

        __VDK_PYLIB_PYTHON_GE__ = 3.6
        __VDK_PYLIB_PYTHON_LT__ = 3.10

    means that the code is compatible only with Python 3.6, 3.7,
    3.8 and 3.9.

    Only Python 3 versions are supported.
